@@ -25,12 +25,12 @@ msgstr ""
 """
 
 with open(sys.argv[1],"w") as fout:
+    fout.write(HEADER)
     for fn in sys.argv[2:]:
-        fout.write(HEADER)
         with open(fn) as f:
             fn = re.sub('.*/dat', 'dat', fn)
             for i, line in enumerate(f):
                 line = line.rstrip('\r\n')
                 if line and not line.startswith('['):
                     quoted_escaped_line = json.dumps(line, ensure_ascii=False)
-                    fout.write('#: {}:{}\nmsgid {}\nmsgstr ""\n'.format(fn, i+1, quoted_escaped_line))
+                    fout.write('#: {}:{}\nmsgid {}\nmsgstr ""\n\n'.format(fn, i+1, quoted_escaped_line))
